// Lista simplemente enlazada

template <class TYPE> class CSList;

#define ASSERT(A) if(!A) throw

class CBaseNode
{
	CBaseNode* m_pNext;
protected:
	CBaseNode() { m_pNext = NULL; }
	CBaseNode(CBaseNode* p) { m_pNext = p; }

	friend class CSinglyLinkedList;
};

typedef CBaseNode* NODEPOS;

class CSinglyLinkedList
{
	NODEPOS m_pLast;		// m_pLast->m_pNext is list head
	int m_nCount;			// Just for empty. I must fix it

protected:
	CSinglyLinkedList()
	{
		m_pLast = NULL;
		m_nCount = 0;
	}

	void Empty()
	{	
		/*NODEPOS pos = GetHeadPos();

		while (pos) 
			delete GetNext(pos);*/
		
		// This implementation doesn't delets lists nodes
		m_pLast = NULL;
		m_nCount = 0;
	}

	void AddHead(NODEPOS pBN)
	{
		if (m_pLast)
			pBN->m_pNext = m_pLast->m_pNext;
		else
			m_pLast = pBN;

		m_pLast->m_pNext = pBN;
		m_nCount++;
	}

	void AddTail(NODEPOS pBN)
	{
		if (m_pLast) {
			pBN->m_pNext = m_pLast->m_pNext;
			m_pLast = m_pLast->m_pNext = pBN;
		}
		else
			m_pLast = pBN->m_pNext = pBN;
					
		m_nCount++;
	}

	void AddTail(CSinglyLinkedList& list)
	{
		NODEPOS headpos = list.GetHeadPos();
		NODEPOS tailpos = list.GetTailPos();
		
		if (headpos) {
			if (m_pLast) {
				tailpos->m_pNext = m_pLast->m_pNext;
				m_pLast->m_pNext = headpos;
			}
			m_pLast = tailpos;

			m_nCount += list.m_nCount;
			
			// list loses its members
			list.m_pLast = NULL;
			list.m_nCount = 0;
		}
	}

	NODEPOS RemoveHead()
	{
		ASSERT(m_pLast);

		NODEPOS p = m_pLast->m_pNext;

		if (p == m_pLast)
			m_pLast = NULL;
		else
			m_pLast->m_pNext = p->m_pNext;

		m_nCount--;

		return p;
	}

	// RemoveTail is no defined yet

public:
	NODEPOS GetHeadPos() const
	{
		if (m_pLast)
			return m_pLast->m_pNext;
		else
			return NULL;
	}

	NODEPOS GetTailPos() const
	{
		return m_pLast;
	}

	bool IsEmpty() const
	{
		return (m_pLast == NULL);
	}

	int GetCount() const
	{
		return m_nCount;
	}

	NODEPOS GetNext(NODEPOS& r)
	{
		ASSERT(r);
		NODEPOS p = r;
		
		if ( r == m_pLast)
			r = NULL;
		else
			r = p->m_pNext;

		return p;
	}
};

template <class TYPE> class CNode : public CBaseNode
{
	TYPE	m_data;

public:
	CNode(const TYPE& data) : m_data(data) { }
	TYPE& GetData() { return m_data; }
};


template <class TYPE> class CSList : public CSinglyLinkedList
{
public:
	~CSList()
	{
		Empty();  // Never call Empty()'s base class
	}
	
	void Empty()
	{
		NODEPOS pos = GetHeadPos();

		while (pos) 
			delete (CNode<TYPE>*)CSinglyLinkedList::GetNext(pos);

		CSinglyLinkedList::Empty();
	}

	void AddHead(const TYPE& data)
	{
		CSinglyLinkedList::AddHead(new CNode<TYPE>(data));
	}

	void AddTail(const TYPE& data)
	{
		CSinglyLinkedList::AddTail(new CNode<TYPE>(data));
	}

	void AddTail(CSList<TYPE>& list)
	{
		CSinglyLinkedList::AddTail(list);
	}
	
	TYPE& GetAt(NODEPOS p)
	{
		ASSERT(p);
				
		return ((CNode<TYPE>*)p)->GetData();
	}

	TYPE& GetNext(NODEPOS& p)
	{
		ASSERT(p);
				
		return ((CNode<TYPE>*)CSinglyLinkedList::GetNext(p))->GetData();
	}

	NODEPOS Find(const TYPE& data/*FindTextFunc FindText, char* sz*/)
	{
		NODEPOS pos = GetHeadPos();

		while (pos)
			if (data == GetAt(pos))
				return pos;
			else
				CSinglyLinkedList::GetNext(pos);
		
		return NULL;
	}
};

template <class TYPE> class CStack : private CSinglyLinkedList
{
public:
	void Push(const TYPE& data)
	{
		CSinglyLinkedList::AddHead(new CNode<TYPE>(data));
	}
	
	TYPE Pop()
	{
		CNode<TYPE>* p = (CNode<TYPE>*)CSinglyLinkedList::RemoveHead();
		TYPE ret = p->GetData();
		
		delete p;

		return ret;
	}

	bool IsEmpty()
	{
		return CSinglyLinkedList::IsEmpty();
	}
};

/* Example
template<class TYPE, class ARG_TYPE>
inline int CList<TYPE, ARG_TYPE>::GetCount() const
	{ return m_nCount; }
*/