%{
	/* Analizador gramatical, TP2 y TP3
	 * Integrantes: Yanina Manassa, Analia Jahnig, Diego Macrini
	 */           
	 #include <windows.h>
	 #include <stdio.h>     
	 #include "SharedTypes.h"
	 
	 extern "C" FILE *yyout;
	 extern int nLine;
	 extern CSList<SYMBOL> symTable;
	 
	 double gNroVersion = 2.10;
	 
	 CSList<CUARTETO> cuartetos;	// Lista de cuartetos
	 CStack<PSYMBOL>  freeVarTemps;	// Pila de variables temporables libres
	 
	 #define ConcatenarCodigo(A, B) A.GetList().AddTail(B.GetList());
	 
	 // Utilizamos nombres distintos para AddCuarteto en lugar de
	 // sobrecargar la funcin para que quede claro que ace c/u.
	 void AddCuartetoOpBinario(int label, SA& arg1, SA& arg2, SA& res);
	 void AddCuartetoOpUnario(int label, PSYMBOL arg1, SA& arg2, SA& res);
	 
	 void ErrorMessage(char* szMsg);
%}

%union {
	SA  	sa;
	PSYMBOL pSymbol;
	int		nConst;
	int 	nCompCode;

	tagYYSTYPE()
	{
		/* Idem constructor de la lista */
		sa.m_pLast = NULL;
		sa.m_nCount = 0;
		
		//ErrorMessage("Se creo la union");
	}
}       
       
%token KW_IF KW_ELSE KW_WHILE KW_SWITCH KW_CASE STRING KW_DEFAULT
%token MENIG MAYIG IGIG DIST

%token <pSymbol>   ID
%token <nConst> CONS 
%token <nCompCode> COMP

%type <sa> programa sentencia expresion lista_sentencias 
%type <sa> condicion condicion_logica condicion_AND condicion_OR

%left '-' '+'
%left '*' '/'
/*%nonassoc UMENOS*/

%%
programa:   sentencia { cuartetos.AddTail($1.GetList()); }
	|		programa sentencia { cuartetos.AddTail($2.GetList()); }
	;

sentencia: 	ID '=' expresion ';'	{
				AddCuartetoOpUnario('=', $1, $3, $$);
				fprintf(yyout, "Sentencia de asignacin valida\n");
			}
	|		ID '=' STRING ';' {
				fprintf(yyout, "Sentencia de asignacin valida del string\n");
			}
	|		expresion ';' {
				ConcatenarCodigo($$, $1);
				fprintf(yyout, "Sentencia valida\n");
			}
/*	|		s_IF {
				fprintf(yyout, "Sentencia if valida\n");
			}*/
	|		s_while	{
				fprintf(yyout, "Sentencia while valida\n");
			}
	|		switch_case { 
				fprintf(yyout, "Sentencia case valida\n");
			}
	|		IF_ELSE	{
				fprintf(yyout, "Sentencia if else valida\n");
			}
	| 		'{' lista_sentencias '}'
	;   
	

                                                                 
s_while:	KW_WHILE '(' condicion ')' sentencia
	|		KW_WHILE '(' condicion ')' '{' lista_sentencias '}'                                                                 
	;
	                                                           
IF_ELSE: 	s_IF KW_ELSE sentencia 
	|		s_IF KW_ELSE '{' lista_sentencias '}'
	;
	
s_IF:		KW_IF '(' condicion ')' sentencia 
	|		KW_IF '(' condicion ')' '{' lista_sentencias '}'
	;
	
lista_sentencias: sentencia { 
				ConcatenarCodigo($$, $1); 
			}
	|		lista_sentencias sentencia { 
				ConcatenarCodigo($$, $1);
				ConcatenarCodigo($$, $2);
			}
	;

switch_case:	KW_SWITCH '(' ID ')' '{' lista_case '}' 
	|			KW_SWITCH '(' ID ')' '{' lista_case s_default '}' 
	;

lista_case:	s_case
	|		lista_case s_case
	;    
     
s_case:		KW_CASE CONS ':' sentencia 
	|		KW_CASE CONS ':' '{' lista_sentencias '}'
	;
	
s_default:	KW_DEFAULT ':' sentencia 
	| 		KW_DEFAULT ':' '{' lista_sentencias '}'
	;
	
condicion:	condicion_logica { 
				$$.value = $1.value;
				$$.nType = $1.nType;
				ConcatenarCodigo($$, $1); 
			}
	|		condicion_AND
	|		condicion_OR
	;
	
condicion_AND:	condicion_logica '&' '&' condicion_logica
	;                                                    
	
condicion_OR:  	condicion_logica '|' '|' condicion_logica
	;
	     
condicion_logica:  	expresion COMP expresion { AddCuartetoOpBinario($2, $1, $3, $$); }
	;

expresion:	expresion '+' expresion { AddCuartetoOpBinario('+',$1,$3,$$); }
	|		expresion '-' expresion { AddCuartetoOpBinario('-',$1,$3,$$); }
	|		expresion '*' expresion { AddCuartetoOpBinario('*',$1,$3,$$); }
	|		expresion '/' expresion { AddCuartetoOpBinario('/',$1,$3,$$); }
	
	/*|		'-' expresion %prec UMENOS  */
	|		'(' expresion ')' { $$ = $2; }  // Ojo aca
	
	|		ID { 
				$$.value.pSymbol = $1; 
				$$.nType = TYPE_SYMBOL;
			}                             
			
	|		CONS { 
				$$.value.nConst = $1; 
				$$.nType = TYPE_CONST;
			} 
	; 

%%
extern "C" FILE *yyin;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
				   PSTR szCmdLine, int iCmdShow)
{
	if (__argc != 3) 
	{
		ErrorMessage("Error en la lnea de comando");
		return 1;
	}
	if( (yyin  = fopen(__argv[1], "rt" )) == NULL ) 
	{
		ErrorMessage("Error al abrir el archivo de entrada");
		return 1;
	}
	if((yyout = fopen(__argv[2], "wt" )) == NULL ) 
	{
		fclose(yyin);
		ErrorMessage("Error al abrir el archivo de salida");
		return 1;
	}
	
	fprintf(yyout, "Inicio del parsing ( versin %.2f )\n\n", gNroVersion);

	while(!feof(yyin)) {
		yyparse();
	}
	
	fprintf(yyout, "\nTabla de smbolos\n\n");
	NODEPOS pos = symTable.GetHeadPos();
	
	while (pos)
		symTable.GetNext(pos).Write(yyout);
		
	fprintf(yyout, "\nCuartetos\n\n");	
	
	pos = cuartetos.GetHeadPos();
	
	while (pos)
		cuartetos.GetNext(pos).Write(yyout, symTable);

	 _fcloseall( );

	return 0;
}

void yyerror(const char* s)
{
	int c;
	if (feof(yyin))
		fprintf(yyout, "Error por fin de archivo");
	else
		fprintf(yyout, "Error de sintaxis en lnea: %d\n", nLine);
	
	do {
		c = getc(yyin);
		if (c == '\n') nLine++;
	} while (c != ';' && c != EOF);
}

void ErrorMessage(char* szMsg)
{
	MessageBox(NULL, szMsg, "Error", MB_OK | MB_ICONERROR);
}

PSYMBOL GetVarTemp()
{
	static i = 0;
	
	if (freeVarTemps.IsEmpty()) {
		SYMBOL sym;
		
		strcpy(sym.name, "_Tmp");
		_itoa(i++, sym.name + 4, 10);
		
		symTable.AddTail(sym);
		
		return symTable.GetTailPos();
	}
	else
		return freeVarTemps.Pop();
}

void FreeVarTemp(PSYMBOL pSymbol)
{
	freeVarTemps.Push(pSymbol);
}

void AddCuartetoOpBinario(int label, SA& arg1, SA& arg2, SA& res)
{	
	ConcatenarCodigo(res, arg1);
	ConcatenarCodigo(res, arg2);
	
	// Si arg1 o arg2 son variables temporales, las liberamos
	// para que puedan se reutilizadas.
	if (arg1.nType & TYPE_TEMP)
		FreeVarTemp(arg1.value.pSymbol);
		
	if (arg2.nType & TYPE_TEMP)
		FreeVarTemp(arg2.value.pSymbol);
	

	res.value.pSymbol = GetVarTemp();
	res.nType = TYPE_SYMBOL | TYPE_TEMP;
	
	CUARTETO c;
	
	c.label = label;

	c.disc = (arg1.nType & TYPE_SYMBOL) ? SYM1:0;
	if (arg2.nType & TYPE_SYMBOL) c.disc |= SYM2;

	c.arg1 = arg1.value;
	c.arg2 = arg2.value;
	c.res = res.value.pSymbol;
	
	res.GetList().AddTail(c);
}

void AddCuartetoOpUnario(int label, PSYMBOL arg1, SA& arg2, SA& res)
{
	ConcatenarCodigo(res, arg2);
	
	// Si arg2 es una variable temporal, la liberamos.
	if (arg2.nType & TYPE_TEMP)
		FreeVarTemp(arg2.value.pSymbol);
	
	// El resultado debe quedar en arg1	                                  
	res.value.pSymbol = arg1;
	
	CUARTETO c;
	
	c.label = label;

	// arg2 va a pasar a ser el arg1 del cuarteto y el
	// arg2 del cuarteto no se utiliza por lo que el
	// discriminante queda como que arg2 del cuarteto es CONST.
	c.disc = (arg2.nType & TYPE_SYMBOL) ? SYM1:0;

	// Advertencia: arg2 debe ser el arg1 del cuarteto
	// y el arg2 del cuarteto no se utiliza. Se le asigna
	// cero para que que en un estado mostrable.
	c.arg1 = arg2.value;
	c.arg2.nConst = 0;		// disc dice que es CONST => OK
	c.res = res.value.pSymbol;
	
	res.GetList().AddTail(c);
}
